#!/usr/bin/python

import click
import unittest
import sys
import os

from ConfigParser import ConfigParser

from django.test.runner import DiscoverRunner

import chisubmit.backend.settings

import django
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "chisubmit.backend.settings")
django.setup()

from chisubmit.tests.integration.complete.test_complete1 import CLICompleteWorkflowExtensionsPerTeam
from chisubmit.tests.integration.complete.test_complete2 import CLICompleteWorkflowExtensionsPerStudent
from chisubmit.tests.integration.complete.test_complete3 import CLICompleteWorkflowCancelSubmission


test_suites = {"api": "chisubmit.tests.unit.api",
               "clientlibs": "chisubmit.tests.integration.clientlibs",
               "cli": "chisubmit.tests.integration.cli",
         
               "complete1": CLICompleteWorkflowExtensionsPerTeam,
               "complete2": CLICompleteWorkflowExtensionsPerStudent,
               "complete3": CLICompleteWorkflowCancelSubmission}
         
unit_tests = ["api"]

integration_tests = ["clientlibs", "cli"]

complete_tests = ["complete1", "complete2", "complete3"]
 
all_except_complete = unit_tests + integration_tests

def run_complete_test(runner, name, test_config, git_server, git_staging):
    suite = unittest.TestSuite()
    test_class = test_suites[name]
    for name in unittest.TestLoader().getTestCaseNames(test_class):
        test = test_class(name)
        configure_complete_test(test, test_config, git_server, git_staging)
        suite.addTest(test)    
    runner.run_tests([], extra_tests=suite)

@click.command()
@click.option("--failfast", "-f", is_flag=True)
@click.option("--quiet", "-q", is_flag=True)
@click.option("--verbose", "-v", is_flag=True)
@click.option("--buffer", "-b", is_flag=True)
@click.option("--config", "-c", type=click.File())
@click.option("--git-server", type=str)
@click.option("--git-staging", type=str)
@click.argument('tests', type=str, default="all")
def runtests(failfast, quiet, verbose, buffer, 
             config, git_server, git_staging, 
             tests):
    verbosity = 1
    if quiet:
        verbosity = 0
    if verbose:
        from chisubmit.tests.common import set_cli_verbose
        set_cli_verbose(True)
        verbosity = 2
        
    if config is not None:
        test_config = ConfigParser()
        test_config.readfp(config)
    else:
        test_config = None

    runner = DiscoverRunner(verbosity=verbosity, failfast=failfast)
        
    ran = False
    
    if tests == "all":
        ran = True
        runner.run_tests([test_suites[t] for t in all_except_complete])
                
    if tests in all_except_complete:
        ran = True
        runner.run_tests([test_suites[tests]])
        
    if tests in complete_tests:
        ran = True
        run_complete_test(runner, tests, test_config, git_server, git_staging)
                
    if tests == "complete":
        ran = True
        for test in complete_tests:
            run_complete_test(runner, test, test_config, git_server, git_staging)
        
        
    if not ran:
        #try:
            runner.run_tests([tests])
        #except AttributeError, ae:
        #    print "Unknown test: %s" % tests
    
    
def configure_complete_test(test, config, git_server, git_staging):
    if config is None and (git_server is not None or git_staging is not None):
        print "ERROR: You have specified a Git server or staging server, but have"
        print "       not provided a configuration file."
        sys.exit(1)
        
    if config is None:
        return
    
    if git_server is not None:
        if not config.has_section(git_server):
            print "ERROR: You specified a %s server, but the configuration file" % git_server
            print "       doesn't have a [%s] section" % git_server
        
        connstr = config.get(git_server, "server-connstr")
        apikey = config.get(git_server, "api-key")
        username = config.get(git_server, "username")
        
        test.set_git_server_connstr(connstr)
        test.add_api_key(git_server, apikey)
        test.set_git_server_user(username)
        
    if git_staging is not None:
        if not config.has_section(git_staging):
            print "ERROR: You specified a %s server, but the configuration file" % gitgit_staging_server
            print "       doesn't have a [%s] section" % git_staging
        
        connstr = config.get(git_staging, "staging-connstr")
        apikey = config.get(git_staging, "api-key")
        username = config.get(git_server, "username")
        
        test.set_git_staging_connstr(connstr)
        test.add_api_key(git_staging, apikey)
        test.set_git_staging_user(username)
            
if __name__ == "__main__":
    runtests()
    


from chisubmit.tests.integration.clientlibs import ChisubmitClientLibsTestCase
from chisubmit.tests.common import ALL_USERS
from django.contrib.auth.models import User
from chisubmit.client.exceptions import UnknownObjectException,\
    UnauthorizedException
from rest_framework.authtoken.models import Token

class UserTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users']
    
    def test_get_users(self):
        c = self.get_api_client("admintoken")
        
        users = c.get_users()
        self.assertItemsEqual([u.username for u in users], ALL_USERS)

    
    def test_get_user(self):
        c = self.get_api_client("admintoken")
        
        user = c.get_user("instructor1")
        self.assertEquals(user.username, "instructor1")
        self.assertEquals(user.first_name, "F_instructor1")
        self.assertEquals(user.last_name, "L_instructor1")

    def test_create_user(self):
        c = self.get_api_client("admintoken")
        
        user = c.create_user(username = "student100",
                             first_name = "F_student100",
                             last_name = "L_student100",
                             email = "student100@example.org")
        self.assertEquals(user.username, "student100")
        self.assertEquals(user.first_name, "F_student100")
        self.assertEquals(user.last_name, "L_student100")
        self.assertEquals(user.email, "student100@example.org")
        
        try:
            user_obj = User.objects.get(username="student100")
        except User.DoesNotExist:
            self.fail("User was not added to database")  
            
        self.assertEquals(user_obj.username, "student100")
        self.assertEquals(user_obj.first_name, "F_student100")
        self.assertEquals(user_obj.last_name, "L_student100")
        self.assertEquals(user_obj.email, "student100@example.org")
        
class UserTokenTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users']
    
    def test_get_own_token_as_admin(self):
        c = self.get_api_client("admintoken")
        
        token, created = c.get_user_token()
        self.assertEqual(token, "admintoken")    
        self.assertEqual(created, False)
        
    def test_get_other_token_as_admin(self):
        c = self.get_api_client("admintoken")
        
        token, created = c.get_user_token(username = "instructor1")
        self.assertEqual(token, "instructor1token")    
        self.assertEqual(created, False)        
        
    def test_get_own_token(self):
        c = self.get_api_client("instructor1", password="instructor1")
        
        token, created = c.get_user_token()
        self.assertEqual(token, "instructor1token")    
        self.assertEqual(created, False)         

    def test_create_own_token(self):
        c = self.get_api_client("instructor1", password="instructor1")
        
        token = Token.objects.get(user__username="instructor1")
        self.assertEqual(token.key, "instructor1token")    
        token.delete()
        
        token, created = c.get_user_token()
        self.assertNotEqual(token, "instructor1token")    
        self.assertEqual(created, True)         
        
    def test_reset_own_token(self):
        c = self.get_api_client("instructor1", password="instructor1")
        
        token, created = c.get_user_token(reset=True)
        self.assertNotEqual(token, "instructor1token")    
        self.assertEqual(created, True)                
            
    def test_get_other_token(self):
        c = self.get_api_client("instructor1", password="instructor1")
        
        with self.assertRaises(UnknownObjectException) as cm:
            token, created = c.get_user_token(username="instructor2")   
            
    def test_basic_auth_fail(self):
        c = self.get_api_client("instructor1", password="foobar")
        
        with self.assertRaises(UnauthorizedException) as cm:
            token, created = c.get_user_token(username="instructor1")                         



from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.contrib.auth.models import User
from chisubmit.common.utils import get_datetime_now_utc
from datetime import timedelta
from chisubmit.backend.api.models import Assignment

class SubmitTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations']  
        
    def test_correct_no_extensions(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions": 0                     
                    }
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        
    def test_correct_dry_run(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions": 0                     
                    }
        response = self.client.post(url + "?dry_run=true", data = post_data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_incorrect_excessive_extensions(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions": 1                     
                    }
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)           
        
    def test_correct_one_extension(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions": 1                   
                    }
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        
    def test_incorrect_insufficient_extensions_requested(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions": 0                
                    }
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        
    def test_incorrect_insufficient_extensions_in_team(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23 + 24 + 24)
        assignment_obj = Assignment.objects.get(assignment_id = "pa1")
        assignment_obj.deadline = deadline
        assignment_obj.save()

        url = reverse('submit', args=["cmsc40100", "student1-student2", "pa1"])
        
        post_data = {
                     "commit_sha": "COMMITSHATEST",
                     "extensions": 3            
                    }
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)                 
        
class SubmitWithExistingSubmissionsTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_teams', 
                         'course1_pa1', 'course1_pa1_registrations_with_submissions']
        
                


from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.contrib.auth.models import User

class RegisterTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users', 'course1_pa1']
        
    def test_register_single_student(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student2"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        
    def test_register_full_team(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student2"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)        

        user = User.objects.get(username='student2')
        self.client.force_authenticate(user=user)

        post_data = {"students": ["student1", "student2"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_register_full_team_redundant(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student2"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)        

        user = User.objects.get(username='student2')
        self.client.force_authenticate(user=user)

        post_data = {"students": ["student1", "student2"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student2"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)        
                
    def test_register_non_student(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student2"]}
        
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        
class RegisterErrorTests(APITestCase):        
        
    fixtures = ['users', 'course1', 'course1_users', 'course1_pa1']
                
    def test_register_other_students(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student2", "student3"]}
        
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        
    def test_register_student_in_other_class(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student5"]}
        
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)        
        
    def test_register_wrong_number_of_students(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student2", "student3"]}
        
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)          
           
class RegisterExistingTeamErrorTests(APITestCase):             
                
    fixtures = ['users', 'course1', 'course1_users', 'course1_pa1', 'course1_teams']
    
    def test_register_in_different_groups(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('register', args=["cmsc40100", "pa1"])
        
        post_data = {"students": ["student1", "student2"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)          
        
        user = User.objects.get(username='student2')
        self.client.force_authenticate(user=user)

        post_data = {"students": ["student2", "student3"]}
        response = self.client.post(url, data = post_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)     
             
                


import click

import os.path
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.cli.common import create_grading_repos,\
    gradingrepo_push_grading_branch, gradingrepo_pull_grading_branch,\
    get_grader_or_exit, get_assignment_or_exit, get_teams_registrations,\
    catch_chisubmit_exceptions, require_local_config
from chisubmit.repos.grading import GradingGitRepo
from chisubmit.rubric import RubricFile, ChisubmitRubricException
from chisubmit.cli.common import pass_course
from chisubmit.cli.shared.course import shared_course_get_git_credentials


@click.group()
@click.pass_context
def grader(ctx):
    pass

@click.command(name="create-local-grading-repos")
@click.option('--grader', type=str)
@click.argument('assignment_id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def grader_create_local_grading_repos(ctx, course, grader, assignment_id):
    if grader is None:
        user = ctx.obj["client"].get_user()    
        
        grader = get_grader_or_exit(ctx, course, user.username)
    else:
        grader = get_grader_or_exit(ctx, course, grader)
        
    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader)
    
    if len(teams_registrations) == 0:
        print "No teams found"
        ctx.exit(CHISUBMIT_FAIL)

    repos = create_grading_repos(ctx.obj['config'], course, assignment, teams_registrations)

    for repo in repos:
        repo.set_grader_author()

    for team, registration in teams_registrations.items():
        print ("Pulling grading branch for team %s... " % team.team_id),
        gradingrepo_pull_grading_branch(ctx.obj['config'], course, team, registration, from_staging = True)
        print "done"
        
    return CHISUBMIT_SUCCESS

def validate_repo_rubric(ctx, course, assignment, team, registration):
    repo = GradingGitRepo.get_grading_repo(ctx.obj['config'], course, team, registration)
    if not repo:
        print "Repository for %s does not exist" % (team.team_id)
        ctx.exit(CHISUBMIT_FAIL)

    rubricfile = repo.repo_path + "/%s.rubric.txt" % assignment.assignment_id

    if not os.path.exists(rubricfile):
        print "Repository for %s does not exist have a rubric for assignment %s" % (team.team_id, assignment.assignment_id)
        ctx.exit(CHISUBMIT_FAIL)

    try:
        RubricFile.from_file(open(rubricfile), assignment)
        return (True, None)
    except ChisubmitRubricException, cre:
        return (False, cre.message)


@click.command(name="validate-rubrics")
@click.argument('assignment_id', type=str)
@click.option('--grader', type=str)
@click.option('--only', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def grader_validate_rubrics(ctx, course, assignment_id, grader, only):
    if grader is None:
        user = ctx.obj["client"].get_user()    
        
        grader = get_grader_or_exit(ctx, course, user.username)
    else:
        grader = get_grader_or_exit(ctx, course, grader)

    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader, only = only)
    
    for team, registration in teams_registrations.items():
        valid, error_msg = validate_repo_rubric(ctx, course, assignment, team, registration)

        if valid:
            print "%s: Rubric OK." % team.team_id
        else:
            print "%s: Rubric ERROR: %s" % (team.team_id, error_msg)

    return CHISUBMIT_SUCCESS


@click.command(name="push-grading-branches")
@click.argument('assignment_id', type=str)
@click.option('--grader', type=str)
@click.option('--only', type=str)
@click.option('--skip-rubric-validation', is_flag=True)
@require_local_config
@pass_course
@click.pass_context
def grader_push_grading_branches(ctx, course, assignment_id, grader, only, skip_rubric_validation):
    if grader is None:
        user = ctx.obj["client"].get_user()    
        
        grader = get_grader_or_exit(ctx, course, user.username)
    else:
        grader = get_grader_or_exit(ctx, course, grader)

    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader, only = only, only_ready_for_grading=True)
    
    if len(teams_registrations) == 0:
        print "No teams found"
        ctx.exit(CHISUBMIT_FAIL)

    for team, registration in teams_registrations.items():
        if not skip_rubric_validation:
            valid, error_msg = validate_repo_rubric(ctx, course, assignment, team, registration)
            if not valid:
                print "Not pushing branch for team %s. Rubric does not validate: %s" % (team.team_id, error_msg)
                continue
        
        print "Pushing grading branch for team %s... " % team.team_id
        gradingrepo_push_grading_branch(ctx.obj['config'], course, team, registration, to_staging = True)

    return CHISUBMIT_SUCCESS

@click.command(name="pull-grading-branches")
@click.argument('assignment_id', type=str)
@click.option('--grader', type=str)
@click.option('--only', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def grader_pull_grading_branches(ctx, course, assignment_id, grader, only):
    if grader is None:
        user = ctx.obj["client"].get_user()    
        
        grader = get_grader_or_exit(ctx, course, user.username)
    else:
        grader = get_grader_or_exit(ctx, course, grader)

    assignment = get_assignment_or_exit(ctx, course, assignment_id)

    teams_registrations = get_teams_registrations(course, assignment, grader = grader, only = only, only_ready_for_grading=True)
    
    if len(teams_registrations) == 0:
        print "No teams found"
        ctx.exit(CHISUBMIT_FAIL)

    for team, registration in teams_registrations.items():
        print "Pulling grading branch for team %s... " % team.team_id
        gradingrepo_pull_grading_branch(ctx.obj['config'], course, team, registration, from_staging = True)

    return CHISUBMIT_SUCCESS

grader.add_command(grader_create_local_grading_repos)
grader.add_command(grader_validate_rubrics)
grader.add_command(grader_push_grading_branches)
grader.add_command(grader_pull_grading_branches)
grader.add_command(shared_course_get_git_credentials)




import click
from chisubmit.cli.common import pass_course, catch_chisubmit_exceptions,\
    require_local_config
from chisubmit.common import CHISUBMIT_FAIL, CHISUBMIT_SUCCESS
from chisubmit.cli.shared.course import shared_course_set_user_attribute,\
    shared_course_get_git_credentials, shared_course_list
    
@click.group(name="course")
@click.pass_context
def instructor_course(ctx):
    pass

@click.command(name="grader-add-conflict")
@click.argument('grader_id', type=str)
@click.argument('student_id', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def instructor_grader_add_conflict(ctx, course, grader_id, student_id):
    grader = course.get_grader(grader_id)
    if grader is None:
        print "Grader %s does not exist" % grader_id
        ctx.exit(CHISUBMIT_FAIL)

    student = course.get_student(student_id)
    if student is None:
        print "Student %s does not exist" % student_id
        ctx.exit(CHISUBMIT_FAIL)

    course.add_grader_conflict(grader, student_id)

    return CHISUBMIT_SUCCESS

instructor_course.add_command(shared_course_list)
instructor_course.add_command(shared_course_set_user_attribute)
instructor_course.add_command(shared_course_get_git_credentials)




#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

from chisubmit.client.types import ChisubmitAPIObject, Attribute, APIStringType,\
    APIIntegerType, APIBooleanType, APIObjectType, APIDateTimeType, APIDictType,\
    APIDecimalType
from chisubmit.client.users import Student, User, Grader
from chisubmit.client.assignment import Assignment, RubricComponent
from chisubmit.common.utils import is_submission_ready_for_grading


class Team(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),

                       "students_url": Attribute(name="students_url", 
                                                 attrtype=APIStringType, 
                                                 editable=False),  
                       
                       "assignments_url": Attribute(name="assignments_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),  
                       
                       "team_id": Attribute(name="team_id", 
                                            attrtype=APIStringType, 
                                            editable=True),  
    
                       "extensions": Attribute(name="extensions", 
                                               attrtype=APIIntegerType, 
                                               editable=True),  
     
                       "active": Attribute(name="active", 
                                           attrtype=APIBooleanType, 
                                           editable=True),
                      }
    
    def get_team_members(self):
        """
        :calls: GET /courses/:course/teams/:team/students/
        :rtype: List of :class:`chisubmit.client.team.TeamMember`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.students_url
        )
        return [TeamMember(self._api_client, headers, elem) for elem in data]        
    
    def get_team_member(self, username):
        """
        :calls: GET /courses/:course/teams/:team/students/:username
        :rtype: :class:`chisubmit.client.team.TeamMember`
        """
        
        assert isinstance(username, (str, unicode)), username
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.students_url + username
        )
        return TeamMember(self._api_client, headers, data)      
    
    def add_team_member(self, user_or_username, confirmed = None):
        """
        :calls: POST /courses/:course/teams/:team/students/
        :rtype: :class:`chisubmit.client.team.TeamMember`
        """
        
        assert isinstance(user_or_username, (str, unicode)) or isinstance(user_or_username, User) 
        
        if isinstance(user_or_username, (str, unicode)):
            username = user_or_username
        elif isinstance(user_or_username, User):
            username = user_or_username.username
        
        post_data = {"username": username}
        
        if confirmed is not None:
            post_data["confirmed"] = confirmed
        
        headers, data = self._api_client._requester.request(
            "POST",
            self.students_url,
            data = post_data
        )
        return TeamMember(self._api_client, headers, data)         
    
    def get_assignment_registrations(self):
        """
        :calls: GET /courses/:course/teams/:team/assignments/
        :rtype: List of :class:`chisubmit.client.team.Registration`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.assignments_url
        )
        return [Registration(self._api_client, headers, elem) for elem in data]        
    
    def get_assignment_registration(self, assignment_id):
        """
        :calls: GET /courses/:course/teams/:team/assignments/:assignment
        :rtype: :class:`chisubmit.client.team.Registration`
        """
        
        assert isinstance(assignment_id, (str, unicode)), assignment_id
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.assignments_url + assignment_id
        )
        return Registration(self._api_client, headers, data)     
      
    def add_assignment_registration(self, assignment_or_assignment_id, grader_or_grader_username = None):
        """
        :calls: POST /courses/:course/teams/:team/assignments/
        :rtype: :class:`chisubmit.client.team.Registration`
        """
        
        assert isinstance(assignment_or_assignment_id, (str, unicode)) or isinstance(assignment_or_assignment_id, Assignment) 
        assert grader_or_grader_username is None or isinstance(grader_or_grader_username, (str, unicode)) or isinstance(grader_or_grader_username, Grader) 
        
        if isinstance(assignment_or_assignment_id, (str, unicode)):
            assignment_id = assignment_or_assignment_id
        elif isinstance(assignment_or_assignment_id, User):
            assignment_id = assignment_or_assignment_id.assignment_id
        
        post_data = {"assignment_id": assignment_id}
        
        if grader_or_grader_username is not None:
            if isinstance(grader_or_grader_username, (str, unicode)):
                grader_username = grader_or_grader_username
            elif isinstance(grader_or_grader_username, User):
                grader_username = grader_or_grader_username.user.username
            post_data["grader_username"] = grader_username

        headers, data = self._api_client._requester.request(
            "POST",
            self.assignments_url,
            data = post_data
        )
        return Registration(self._api_client, headers, data)         
    
class TeamMember(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),

                       "username": Attribute(name="username", 
                                            attrtype=APIStringType, 
                                            editable=False),  
    
                       "student": Attribute(name="student", 
                                            attrtype=APIObjectType(Student), 
                                            editable=False),  
                       
                       "confirmed": Attribute(name="confirmed", 
                                              attrtype=APIBooleanType, 
                                              editable=True),                       
                      }
    
class Submission(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),

                       "id": Attribute(name="id", 
                                       attrtype=APIIntegerType, 
                                       editable=False),  
    
                       "extensions_used": Attribute(name="extensions_used", 
                                               attrtype=APIIntegerType, 
                                               editable=True),  

                       "commit_sha": Attribute(name="commit_sha", 
                                            attrtype=APIStringType, 
                                            editable=True),  
    
                       "submitted_at": Attribute(name="submitted_at", 
                                        attrtype=APIDateTimeType, 
                                        editable=True)                    
                      }          
    
class Registration(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),

                       "assignment_id": Attribute(name="assignment_id", 
                                                  attrtype=APIStringType, 
                                                  editable=False),  
    
                       "assignment": Attribute(name="assignment", 
                                               attrtype=APIObjectType(Assignment), 
                                               editable=False),  

                       "grader_username": Attribute(name="grader_username", 
                                                    attrtype=APIStringType, 
                                                    editable=True),  
    
                       "grader": Attribute(name="grader", 
                                        attrtype=APIObjectType(Grader), 
                                        editable=False),  
                       
                       "submissions_url": Attribute(name="submissions_url", 
                                                    attrtype=APIStringType, 
                                                    editable=False),     
                                              
                       "final_submission_id": Attribute(name="final_submission_id", 
                                                     attrtype=APIIntegerType, 
                                                     editable=True),

                       "final_submission": Attribute(name="final_submission", 
                                                     attrtype=APIObjectType(Submission), 
                                                     editable=False),
                       
                       "final_submission_url": Attribute(name="final_submission_url", 
                                                         attrtype=APIStringType, 
                                                         editable=False),                                     

                       "grades_url": Attribute(name="grades_url", 
                                               attrtype=APIStringType, 
                                               editable=False),                                     

                       "grade_adjustments": Attribute(name="grade_adjustments", 
                                                      attrtype=APIDictType(APIDecimalType), 
                                                      editable=True),                                     
                      }
    
    def get_submissions(self):
        """
        :calls: GET /courses/:course/teams/:team/assignments/:assignment/submissions
        :rtype: List of :class:`chisubmit.client.team.Submission`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.submissions_url
        )
        return [Submission(self._api_client, headers, elem) for elem in data]        
    
    def get_submission(self, submission):
        """
        :calls: GET /courses/:course/teams/:team/assignments/:assignment/submissions/:submission
        :rtype: :class:`chisubmit.client.team.Submission`
        """
        
        assert isinstance(submission, int), submission
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.submissions_url + str(submission)
        )
        return Submission(self._api_client, headers, data)      
    
    def add_submission(self, commit_sha, extensions_used = None, submitted_at = None):
        """
        :calls: POST /courses/:course/teams/:team/assignments/:assignment/submissions/
        :rtype: :class:`chisubmit.client.team.Submission`
        """
        
        post_data = {"commit_sha": commit_sha}
        
        if extensions_used is not None:
            post_data["extensions_used"] = extensions_used
        if submitted_at is not None:
            post_data["submitted_at"] = submitted_at

        headers, data = self._api_client._requester.request(
            "POST",
            self.submissions_url,
            data = post_data
        )
        return Submission(self._api_client, headers, data)

    def get_grades(self):
        """
        :calls: GET /courses/:course/teams/:team/assignments/:assignment/grades/
        :rtype: List of :class:`chisubmit.client.team.Grade`
        """
        
        headers, data = self._api_client._requester.request(
            "GET",
            self.grades_url
        )
        return [Grade(self._api_client, headers, elem) for elem in data]            
    
    def add_grade(self, rubric_component, points = None):
        """
        :calls: POST /courses/:course/teams/:team/assignments/:assignment/grades/
        :rtype: :class:`chisubmit.client.team.Grade`
        """
        
        post_data = {"rubric_component_id": rubric_component.id}
        
        if points is not None:
            post_data["points"] = points

        headers, data = self._api_client._requester.request(
            "POST",
            self.grades_url,
            data = post_data
        )
        return Grade(self._api_client, headers, data)

    
    def submit(self, commit_sha, extensions, ignore_deadline = False, dry_run=False):
        """
        :calls: POST /courses/:course/teams/:team/assignments/:assignment/submit/
        :rtype: :class:`chisubmit.client.team.Submission`
        """
        
        post_data = {"commit_sha": commit_sha,
                     "extensions": extensions,
                     "ignore_deadline": ignore_deadline}
        
        if dry_run:
            qs = "?dry_run=true"
        else:
            qs = ""
        
        headers, data = self._api_client._requester.request(
            "POST",
            self.url + "/submit"+qs,
            data = post_data
        )
        return SubmissionResponse(self._api_client, headers, data)    

    def set_grade(self, rubric_component, points):
        if points < 0 or points > rubric_component.points:
            raise ValueError("Invalid grade value %.2f ('%s' must be 0 <= x <= %.2f)" % (points, rubric_component.description, rubric_component.points))
        
        grades = self.get_grades()
        grade = [g for g in grades if g.rubric_component_id == rubric_component.id]
        
        if len(grade) == 0:
            self.add_grade(rubric_component, points)
        elif len(grade) == 1:
            grade[0].points = points
        else:
            msg = "Server returned more than one grade for '%s' in %s. " % (rubric_component.description, self.assignment.assignment_id)
            msg += "This should not happen. Please contact the chisubmit administrator."

            raise Exception(msg)

    def get_total_penalties(self):
        if self.grade_adjustments is None:
            return 0.0
        else:
            return sum([v for v in self.grade_adjustments.values() if v < 0.0])

    def get_total_bonuses(self):
        if self.grade_adjustments is None:
            return 0.0
        else:
            return sum([v for v in self.grade_adjustments.values() if v >= 0.0])
        
    def get_total_adjustments(self):
        if self.grade_adjustments is None:
            return 0.0
        else:
            return sum([v for v in self.grade_adjustments.values()])
    
    def get_total_grade(self):
        grades = self.get_grades()
        
        return sum([g.points for g in grades]) + self.get_total_adjustments()

    def get_grading_branch_name(self):
        return self.assignment.assignment_id + "-grading"
        
    def is_ready_for_grading(self):    
        if self.final_submission is None:
            return False
        else:
            return is_submission_ready_for_grading(assignment_deadline=self.assignment.deadline, 
                                                   submission_date=self.final_submission.submitted_at,
                                                   extensions_used=self.final_submission.extensions_used)    
    
class SubmissionResponse(ChisubmitAPIObject):
    
    _api_attributes = {
                       "submission": Attribute(name="registration", 
                                               attrtype=APIObjectType(Submission), 
                                               editable=False),  
                                              
                       "extensions_before": Attribute(name="extensions_before", 
                                                      attrtype=APIIntegerType, 
                                                      editable=False),  

                       "extensions_after": Attribute(name="extensions_before", 
                                                     attrtype=APIIntegerType, 
                                                     editable=False),  
                       }    
    
    
class Grade(ChisubmitAPIObject):

    _api_attributes = {"url": Attribute(name="url", 
                                       attrtype=APIStringType, 
                                       editable=False),
                       
                       "rubric_component_id": Attribute(name="rubric_component_id", 
                                                        attrtype=APIIntegerType, 
                                                        editable=False),
                             
                       "rubric_component": Attribute(name="rubric_component", 
                                                     attrtype=APIObjectType(RubricComponent), 
                                                     editable=False),   
                       
                       "points": Attribute(name="points", 
                                           attrtype=APIDecimalType, 
                                           editable=True)
                       }                          


from django.db import models
from django.core.validators import MinValueValidator
from django.contrib.auth.models import User
from enum import Enum
from django.db.utils import IntegrityError
from chisubmit.common.utils import compute_extensions_needed,\
    is_submission_ready_for_grading
from rest_framework.response import Response
from rest_framework import status
import jsonfield

class CourseRoles(Enum):
    ADMIN = 0
    INSTRUCTOR = 1
    GRADER = 2
    STUDENT = 3

GradersAndStudents = set([CourseRoles.GRADER, CourseRoles.STUDENT])
AllExceptAdmin = set([CourseRoles.INSTRUCTOR, CourseRoles.GRADER, CourseRoles.STUDENT])
Students = set([CourseRoles.STUDENT])

class OwnerPermissions(Enum):
    READ = 0
    WRITE = 1

Read = set([OwnerPermissions.READ])
Write = set([OwnerPermissions.WRITE])
ReadWrite = Read | Write

def get_user_by_username(username):
    try:
        return User.objects.get(username=username)
    except User.DoesNotExist:
        return None    

class Course(models.Model):
    course_id = models.SlugField(unique = True)
    name = models.CharField(max_length=64)
    
    instructors = models.ManyToManyField(User, through='Instructor', related_name="instructor_in")
    graders = models.ManyToManyField(User, through='Grader', related_name="grader_in")
    students = models.ManyToManyField(User, through='Student', related_name="student_in")
    
    def __unicode__(self):
        return u"%s: %s" % (self.course_id, self.name)
    
    @classmethod
    def get_by_course_id(cls, course_id):
        try:
            return cls.objects.get(course_id=course_id)
        except cls.DoesNotExist:
            return None
            
    def has_instructor(self, user):
        return Instructor.objects.filter(course=self, user=user).exists()
    
    def get_instructor(self, user):
        try:
            return Instructor.objects.get(course=self, user=user)
        except Instructor.DoesNotExist:
            return None        

    def has_grader(self, user):
        return Grader.objects.filter(course=self, user=user).exists()

    def get_grader(self, user):
        try:
            return Grader.objects.get(course=self, user=user)
        except Grader.DoesNotExist:
            return None        

    def has_student(self, user):
        return Student.objects.filter(course=self, user=user).exists()
    
    def get_student(self, user):
        try:
            return Student.objects.get(course=self, user=user)
        except Student.DoesNotExist:
            return None        

    def has_user(self, user):
        return self.has_student(user) or self.has_instructor(user) or self.has_grader(user) 
    
    def has_access(self, user):
        return user.is_staff or user.is_superuser or self.has_user(user)
    
    def get_roles(self, user):
        roles = set()
        if self.has_instructor(user):
            roles.add(CourseRoles.INSTRUCTOR)
        if self.has_grader(user):
            roles.add(CourseRoles.GRADER)
        if self.has_student(user):
            roles.add(CourseRoles.STUDENT)
        if user.is_staff or user.is_superuser:
            roles.add(CourseRoles.ADMIN)
        return roles
    
    def get_assignment(self, assignment_id):
        try:
            return Assignment.objects.get(course__course_id=self.course_id, assignment_id=assignment_id)
        except Assignment.DoesNotExist:
            return None
        
    def get_team(self, team_id):
        try:
            return self.team_set.get(team_id=team_id)
        except Team.DoesNotExist:
            return None        

    def get_teams(self):
        return self.team_set.all()
        
    def get_teams_with_students(self, students):
        return self.team_set.filter(students__in = students).distinct()
    
    # OPTIONS
    GIT_USERNAME_USER = 'user-id'
    GIT_USERNAME_CUSTOM = 'custom'
    GIT_USERNAME_CHOICES = (
        (GIT_USERNAME_USER, 'Same as user id'),
        (GIT_USERNAME_CUSTOM, 'Custom git username'),
    )
    
    EXT_PER_TEAM = "per-team"
    EXT_PER_STUDENT = "per-student"
    EXT_CHOICES = (
        (EXT_PER_TEAM, 'Extensions per team'),
        (EXT_PER_STUDENT, 'Extensions per student'),
    )
    
    git_server_connstr = models.CharField(max_length=256, null=True)
    git_staging_connstr = models.CharField(max_length=256, null=True)
    git_usernames = models.CharField(max_length=16, choices=GIT_USERNAME_CHOICES, default=GIT_USERNAME_USER)
    git_staging_usernames = models.CharField(max_length=16, choices=GIT_USERNAME_CHOICES, default=GIT_USERNAME_USER)
    extension_policy = models.CharField(max_length=16, choices=EXT_CHOICES, default=EXT_PER_STUDENT)
    default_extensions = models.IntegerField(default=0, validators = [MinValueValidator(0)])    
    
class Instructor(models.Model):
    user = models.ForeignKey(User)
    course = models.ForeignKey(Course)
    
    git_username = models.CharField(max_length=64, null=True)
    git_staging_username = models.CharField(max_length=64, null=True)

    def __unicode__(self):
        return u"Instructor %s of %s" % (self.user.username, self.course.course_id) 

    class Meta:
        unique_together = ("user", "course")

class Grader(models.Model):
    user = models.ForeignKey(User)
    course = models.ForeignKey(Course)
    
    git_username = models.CharField(max_length=64)
    git_staging_username = models.CharField(max_length=64)
    
    conflicts = models.ManyToManyField("Student", blank=True)
    
    def __unicode__(self):
        return u"Grader %s of %s" % (self.user.username, self.course.course_id)     
    
    class Meta:
        unique_together = ("user", "course")    
    
class Student(models.Model):
    user = models.ForeignKey(User)
    course = models.ForeignKey(Course)
    
    git_username = models.CharField(max_length=64)
    
    extensions = models.IntegerField(default=0, validators = [MinValueValidator(0)])
    dropped = models.BooleanField(default = False)
    
    def __unicode__(self):
        return u"Student %s of %s" % (self.user.username, self.course.course_id)     
    
    def get_extensions_available(self):
        extensions_used = 0
        
        teams = self.course.get_teams_with_students([self])
        
        for team in teams:
            for registration in team.registration_set.all():
                if registration.final_submission is not None:
                    extensions_used += registration.final_submission.extensions_used

        return self.extensions - extensions_used    
    
    class Meta:
        unique_together = ("user", "course")    

class Assignment(models.Model):
    course = models.ForeignKey(Course)
    assignment_id = models.SlugField()
    name = models.CharField(max_length=64)
    deadline = models.DateTimeField()

    # Options
    min_students = models.IntegerField(default=1, validators = [MinValueValidator(1)])
    max_students = models.IntegerField(default=1, validators = [MinValueValidator(1)])  
            
    def __unicode__(self):
        return u"Assignment %s of %s" % (self.assignment_id, self.course.course_id)     

    def get_rubric_components(self):
        return list(RubricComponent.objects.filter(assignment=self))
    
    def get_rubric_component_by_id(self, rc_id):
        try:
            return RubricComponent.objects.get(pk = rc_id)
        except RubricComponent.DoesNotExist:
            return None

    def get_rubric_component_by_description(self, description):
        try:
            return RubricComponent.objects.get(assignment = self, description = description)
        except RubricComponent.DoesNotExist:
            return None

    class Meta:
        unique_together = ("assignment_id", "course")    

class RubricComponent(models.Model):    
    assignment = models.ForeignKey(Assignment)
    order = models.IntegerField(default=0, validators = [MinValueValidator(0)])
    description = models.CharField(max_length=64)
    points = models.DecimalField(max_digits=5, decimal_places=2)
    
    class Meta:
        unique_together = ("assignment", "description")
        ordering = ("assignment", "order")
        
        
class Team(models.Model):
    course = models.ForeignKey(Course)
    team_id = models.SlugField(max_length=128)
    extensions = models.IntegerField(default=0, validators = [MinValueValidator(0)])
    active = models.BooleanField(default = True)
    
    students = models.ManyToManyField(Student, through='TeamMember', related_name="team_member_in")
    
    registrations = models.ManyToManyField(Assignment, through='Registration') 
    
    def __unicode__(self):
        return u"Team %s in %s" % (self.team_id, self.course.course_id)         
    
    def is_registered_for_assignment(self, assignment):
        return self.registrations.filter(assignment_id = assignment.assignment_id).exists()
    
    def get_registrations(self):
        return self.registration_set.all()    
    
    def get_registration(self, assignment):
        try:
            return self.registration_set.get(assignment = assignment)
        except Registration.DoesNotExist:
            return None        

    def get_team_members(self):
        return self.teammember_set.all()
        
    def get_team_member(self, student):
        try:
            return self.teammember_set.get(student = student)
        except TeamMember.DoesNotExist:
            return None        
        
    def get_extensions_used(self):
        extensions = 0
        for r in self.registration_set.all():
            if r.final_submission is not None:
                extensions += r.final_submission.extensions_used
        return extensions 
        
    def get_extensions_available(self):
        if self.course.extension_policy == Course.EXT_PER_TEAM:
            return self.extensions - self.get_extensions_used()    
        elif self.course.extension_policy == Course.EXT_PER_STUDENT:
            student_extensions_available = []
            for student in self.students.all():
                a = student.get_extensions_available()
                student_extensions_available.append(a)
            return min(student_extensions_available)
        else:
            raise IntegrityError("course.extension_policy has invalid value: %s" % (self.course.extension_policy))          
    
    class Meta:
        unique_together = ("course", "team_id")
        
class TeamMember(models.Model):
    student = models.ForeignKey(Student)
    team = models.ForeignKey(Team)
    
    confirmed = models.BooleanField(default = False)
    
    class Meta:
        unique_together = ("student", "team")        
        
class Registration(models.Model):
    team = models.ForeignKey(Team)
    assignment = models.ForeignKey(Assignment)

    grader = models.ForeignKey(Grader, null=True)
    grade_adjustments = jsonfield.JSONField(blank=True, null=True)
    final_submission = models.ForeignKey("Submission", related_name="final_submission_of", null=True) 

    def is_ready_for_grading(self):
        if self.final_submission is None:
            return False
        else:
            return is_submission_ready_for_grading(assignment_deadline=self.assignment.deadline, 
                                                   submission_date=self.final_submission.submitted_at,
                                                   extensions_used=self.final_submission.extensions_used)

    class Meta:
        unique_together = ("team", "assignment")

class Submission(models.Model):
    registration = models.ForeignKey(Registration) 
    extensions_used = models.IntegerField(default=0, validators = [MinValueValidator(0)])
    commit_sha = models.CharField(max_length=40)
    submitted_at = models.DateTimeField(auto_now_add=True)
    
    def validate(self, ignore_deadline = False):
        extensions_needed = compute_extensions_needed(submission_time = self.submitted_at, 
                                                      deadline = self.registration.assignment.deadline)
        extensions_available = self.registration.team.get_extensions_available()
                
        if extensions_available < 0:
            msg = "The number of available extensions is negative"
            return False, Response({"fatal": [msg]}, status=status.HTTP_500_INTERNAL_SERVER_ERROR), None           

        if self.registration.final_submission is not None:
            extensions_used_in_existing_submission = self.registration.final_submission.extensions_used
        else:
            extensions_used_in_existing_submission = 0 
        
        error_data = {"extensions_available": extensions_available,
                      "extensions_requested": self.extensions_used,
                      "extensions_needed": extensions_needed,
                      "submitted_at": self.submitted_at.isoformat(sep=" "),
                      "deadline": self.registration.assignment.deadline.isoformat(sep=" ")}
        
        if not ignore_deadline and self.extensions_used != extensions_needed:
            msg = "The number of requested extensions does not match the number of extensions needed."
            response_data = {"errors": [msg]}
            response_data.update(error_data)
            return False, Response(response_data, status=status.HTTP_400_BAD_REQUEST), None            
        if not ignore_deadline and extensions_available + extensions_used_in_existing_submission < extensions_needed:
            msg = "The number of available extensions is insufficient."
            response_data = {"errors": [msg]}
            response_data.update(error_data)
            return False, Response(response_data, status=status.HTTP_400_BAD_REQUEST), None                    
        else:     
            extensions = {}
            extensions["extensions_available_before"] = extensions_available
            
            # If the team has already used extensions for a previous submission,
            # they don't count towards the number of extensions needed
            # They are 'credited' to the available extensions
            extensions_available += extensions_used_in_existing_submission
            
            if ignore_deadline:
                extensions_available -= self.extensions_used
            else:
                extensions_available -= extensions_needed

            extensions["extensions_available_after"] = extensions_available
            
            return True, None, extensions                    
        
class Grade(models.Model):
    registration = models.ForeignKey(Registration)
    rubric_component = models.ForeignKey(RubricComponent)
    
    points = models.DecimalField(max_digits=5, decimal_places=2)
    
    class Meta:
        unique_together = ("registration", "rubric_component")    
    


import click
from chisubmit.client.assignment import Assignment
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.common.utils import convert_datetime_to_local,\
    create_connection, get_datetime_now_utc, compute_extensions_needed,\
    is_submission_ready_for_grading
from dateutil.parser import parse
from chisubmit.cli.common import pass_course, get_assignment_or_exit,\
    get_team_or_exit, get_assignment_registration_or_exit,\
    catch_chisubmit_exceptions, require_local_config
from chisubmit.cli.shared.assignment import shared_assignment_list
from datetime import timedelta
from chisubmit.client.exceptions import BadRequestException,\
    UnknownObjectException


@click.group(name="assignment")
@click.pass_context
def student_assignment(ctx):
    pass


@click.command(name="register")
@click.argument('assignment_id', type=str)
@click.option('--partner', type=str, multiple=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def student_assignment_register(ctx, course, assignment_id, partner):
    assignment = get_assignment_or_exit(ctx, course, assignment_id)
    
    user = ctx.obj["client"].get_user()
    
    try:
        course.get_instructor(user.username)
        
        # If get_instructor doesn't raise an exception, then the user
        # is an instructor, and we don't include the user in the list
        # of partners.
        r = assignment.register(students = partner)    
    except UnknownObjectException:
        # Otherwise, we include the current user
        r = assignment.register(students = partner + (user.username,))    

    print "Your registration for %s (%s) is complete." % (r.registration.assignment.assignment_id, r.registration.assignment.name)
    
    if len(r.team_members) > 1:
        some_unconfirmed = False
        print 
        print "Your team name is '%s'." % r.team.team_id
        print
        print "The team is composed of the following students:"
        print
        for tm in r.team_members:
            if tm.confirmed:
                conf = ""
            else:
                conf = "UNCONFIRMED"
                some_unconfirmed = True
                
            print " - %s, %s (%s) %s" % (tm.student.user.last_name, 
                                         tm.student.user.first_name,
                                         tm.student.user.username,
                                         conf)
             
            if some_unconfirmed:
                print
                print "Note: Some students have not yet confirmed that they are part of"
                print "      this team. To confirm they are part of this team, they just"
                print "      need to register as a team themselves (using this same"
                print "      command, and listing the same team members)."
    
    return CHISUBMIT_SUCCESS

@click.command(name="show-deadline")
@click.argument('assignment_id', type=str)
@click.option('--utc', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def student_assignment_show_deadline(ctx, course, assignment_id, utc):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    now_utc = get_datetime_now_utc()
    now_local = convert_datetime_to_local(now_utc)

    deadline_utc = assignment.deadline
    deadline_local = convert_datetime_to_local(deadline_utc)

    print assignment.name
    print
    if utc:
        print "      Now (Local): %s" % now_local.isoformat(" ")
        print " Deadline (Local): %s" % deadline_local.isoformat(" ")
        print
        print "        Now (UTC): %s" % now_utc.isoformat(" ")
        print "   Deadline (UTC): %s" % deadline_utc.isoformat(" ")
    else:
        print "      Now: %s" % now_local.isoformat(" ")
        print " Deadline: %s" % deadline_local.isoformat(" ")

    print

    extensions = compute_extensions_needed(now_utc, deadline_utc)

    if extensions == 0:
        diff = deadline_utc - now_utc
    else:
        diff = now_utc - deadline_utc

    days = diff.days
    hours = diff.seconds // 3600
    minutes = (diff.seconds//60)%60
    seconds = diff.seconds%60

    if extensions == 0:
        print "The deadline has not yet passed"
        print "You have %i days, %i hours, %i minutes, %i seconds left" % (days, hours, minutes, seconds)
    else:
        print "The deadline passed %i days, %i hours, %i minutes, %i seconds ago" % (days, hours, minutes, seconds)
        print "If you submit your assignment now, you will need to use %i extensions" % extensions

    return CHISUBMIT_SUCCESS


def print_commit(commit):
    print "      Commit: %s" % commit.sha
    print "        Date: %s" % commit.committed_date.isoformat(sep=" ")
    print "     Message: %s" % commit.message
    print "      Author: %s <%s>" % (commit.author_name, commit.author_email)        


@click.command(name="submit")
@click.argument('team_id', type=str)    
@click.argument('assignment_id', type=str)
@click.argument('commit_sha', type=str)
@click.option('--extensions', type=int, default=0)
@click.option('--force', is_flag=True)
@click.option('--yes', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context  
def student_assignment_submit(ctx, course, team_id, assignment_id, commit_sha, extensions, force, yes):
    team = get_team_or_exit(ctx, course, team_id)
    registration = get_assignment_registration_or_exit(ctx, team, assignment_id)
    
    if registration.final_submission is not None:
        if is_submission_ready_for_grading(assignment_deadline=registration.assignment.deadline, 
                                           submission_date=registration.final_submission.submitted_at,
                                           extensions_used=registration.final_submission.extensions_used):
            print "You cannot re-submit this assignment."
            print "You made a submission before the deadline, and the deadline has passed."
    
            ctx.exit(CHISUBMIT_FAIL)
        
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
    
    commit = conn.get_commit(course, team, commit_sha)
    
    if commit is None:
        print "Commit %s does not exist in repository" % commit_sha
        ctx.exit(CHISUBMIT_FAIL)

    try:
        submit_response = registration.submit(commit_sha, extensions, dry_run=True)
    except BadRequestException, bre:
        response_data = bre.json
        extensions_needed = response_data["extensions_needed"]
        extensions_available = response_data["extensions_available"]
        
        deadline_utc = parse(response_data["deadline"])
        submitted_at_utc = parse(response_data["submitted_at"])
        deadline_local = convert_datetime_to_local(deadline_utc)
        submitted_at_local = convert_datetime_to_local(submitted_at_utc)        
        
        if extensions_needed > extensions_available:
            msg1 = "You do not have enough extensions to submit this assignment."
            msg2 = "You would need %i extensions to submit this assignment at this " \
                   "time, but you only have %i left" % (extensions_needed, extensions_available)
        elif extensions < extensions_needed:
            msg1 = "The number of extensions you have requested is insufficient."
            msg2 = "You need to request %s extensions." % extensions_needed
        elif extensions > extensions_needed:
            msg1 = "The number of extensions you have requested is excessive."
            msg2 = "You only need to request %s extensions." % extensions_needed

        print
        print msg1
        print            
        print "     Deadline (UTC): %s" % deadline_utc.isoformat(sep=" ")
        print "          Now (UTC): %s" % submitted_at_utc.isoformat(sep=" ")
        print 
        print "   Deadline (Local): %s" % deadline_local.isoformat(sep=" ")
        print "        Now (Local): %s" % submitted_at_local.isoformat(sep=" ")
        print 
        print msg2 
        print

        ctx.exit(CHISUBMIT_FAIL)
    
    if registration.final_submission is not None:
        prior_commit_sha = registration.final_submission.commit_sha
        prior_extensions_used = registration.final_submission.extensions_used             
        prior_submitted_at_utc = registration.final_submission.submitted_at
        prior_submitted_at_local = convert_datetime_to_local(prior_submitted_at_utc)            
        
        submission_commit = conn.get_commit(course, team, prior_commit_sha)            
        
        if prior_commit_sha == commit_sha:
            print "You have already submitted assignment %s" % registration.assignment.assignment_id
            print "You submitted the following commit on %s:" % prior_submitted_at_local
            print
            if submission_commit is None:
                print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
            else:
                print_commit(submission_commit)
            print
            print "You are trying to submit the same commit again (%s)" % prior_commit_sha
            print "If you want to re-submit, please specify a different commit"
            ctx.exit(CHISUBMIT_FAIL)
            
        if not force:
            print        
            print "You have already submitted assignment %s" % registration.assignment.assignment_id
            print "You submitted the following commit on %s:" % prior_submitted_at_local
            print
            if submission_commit is None:
                print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
            else:
                print_commit(submission_commit)
            print
            print "If you want to submit again, please use the --force option"
            ctx.exit(CHISUBMIT_FAIL)
        else:
            print
            print "WARNING: You have already submitted assignment %s and you" % registration.assignment.assignment_id 
            print "are about to overwrite the previous submission of the following commit:"
            print
            if submission_commit is None:
                print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
            else:
                print_commit(submission_commit)
            print

    if registration.final_submission is not None and force:
        msg = "THE ABOVE SUBMISSION FOR %s (%s) WILL BE CANCELLED." % (registration.assignment.assignment_id, registration.assignment.name)
        
        print "!"*len(msg)
        print msg
        print "!"*len(msg)
        print
        print "If you continue, your submission for %s (%s)" % (registration.assignment.assignment_id, registration.assignment.name)
        print "will now point to the following commit:"                
    else:
        print "You are going to make a submission for %s (%s)." % (registration.assignment.assignment_id, registration.assignment.name)
        print "The commit you are submitting is the following:"
    print
    print_commit(commit)
    print
    print "PLEASE VERIFY THIS IS THE EXACT COMMIT YOU WANT TO SUBMIT"
    print
    print "Your team currently has %i extensions" % (submit_response.extensions_before)
    print
    if registration.final_submission is not None:
        print "You used %i extensions in your previous submission of this assignment." % prior_extensions_used
        print "and you are going to use %i additional extensions now." % (extensions - prior_extensions_used)
    else:
        print "You are going to use %i extensions on this submission." % extensions
    print
    print "You will have %i extensions left after this submission." % submit_response.extensions_after
    print 
    print "Are you sure you want to continue? (y/n): ", 
    
    if not yes:
        yesno = raw_input()
    else:
        yesno = 'y'
        print 'y'
    
    if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):
        try:
            submit_response = registration.submit(commit_sha, extensions, dry_run=False)
            
            # TODO: Can't do this until GitLab supports updating tags
            #    
            # message = "Extensions: %i\n" % extensions_requested
            # if submission_tag is None:
            #     conn.create_submission_tag(course, team, tag_name, message, commit.sha)
            # else:
            #     conn.update_submission_tag(course, team, tag_name, message, commit.sha)
            
            print "Your submission has been completed."
            return CHISUBMIT_SUCCESS

        except BadRequestException, bre:        
            print "ERROR: Your submission was not completed. The server reported the following errors:"
            bre.print_errors()
            return CHISUBMIT_FAIL        
    
    
@click.command(name="cancel-submit")
@click.argument('team_id', type=str)    
@click.argument('assignment_id', type=str)
@click.option('--yes', is_flag=True)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context  
def student_assignment_cancel_submit(ctx, course, team_id, assignment_id, yes):
    team = get_team_or_exit(ctx, course, team_id)
    registration = get_assignment_registration_or_exit(ctx, team, assignment_id)
    
    if registration.final_submission is None:
        print "Team %s has not made a submission for assignment %s," % (team_id, assignment_id)
        print "so there is nothing to cancel."
        ctx.exit(CHISUBMIT_FAIL)
        
    if is_submission_ready_for_grading(assignment_deadline=registration.assignment.deadline, 
                                       submission_date=registration.final_submission.submitted_at,
                                       extensions_used=registration.final_submission.extensions_used):
        print "You cannot cancel this submission."
        print "You made a submission before the deadline, and the deadline has passed."

        ctx.exit(CHISUBMIT_FAIL)        
            
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
        
    submission_commit = conn.get_commit(course, team, registration.final_submission.commit_sha)
        
    print
    print "This is your existing submission for assignment %s:" % assignment_id
    print
    if submission_commit is None:
        print "WARNING: Previously submitted commit '%s' is not in the repository!" % registration.final_submission.commit_sha
    else:
        print_commit(submission_commit)
    print    

    print "Are you sure you want to cancel this submission? (y/n): ", 
    
    if not yes:
        yesno = raw_input()
    else:
        yesno = 'y'
        print 'y'
    
    if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):
        registration.final_submission_id = None
        
        # TODO: Can't do this until GitLab supports updating tags
        #    
        # message = "Extensions: %i\n" % extensions_requested
        # if submission_tag is None:
        #     conn.create_submission_tag(course, team, tag_name, message, commit.sha)
        # else:
        #     conn.update_submission_tag(course, team, tag_name, message, commit.sha)
            
        print "Your submission has been cancelled."
        

student_assignment.add_command(shared_assignment_list)

student_assignment.add_command(student_assignment_register)
student_assignment.add_command(student_assignment_show_deadline)
student_assignment.add_command(student_assignment_submit)
student_assignment.add_command(student_assignment_cancel_submit)



from chisubmit.backend.api import views
from rest_framework.urlpatterns import format_suffix_patterns
from django.conf.urls import url

URL_PREFIX = r"^api/v1/"

urlpatterns = [
    url(URL_PREFIX + r'courses/$', views.CourseList.as_view(), name="course-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/$', views.CourseDetail.as_view(), name="course-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/instructors/$', views.InstructorList.as_view(), name="instructor-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/instructors/(?P<username>[a-zA-Z0-9_-]+)$', views.InstructorDetail.as_view(), name="instructor-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/graders/$', views.GraderList.as_view(), name="grader-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/graders/(?P<username>[a-zA-Z0-9_-]+)$', views.GraderDetail.as_view(), name="grader-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/students/$', views.StudentList.as_view(), name="student-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/students/(?P<username>[a-zA-Z0-9_-]+)$', views.StudentDetail.as_view(), name="student-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/assignments/$', views.AssignmentList.as_view(), name="assignment-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)$', views.AssignmentDetail.as_view(), name="assignment-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/rubric/$', views.RubricList.as_view(), name="rubric-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/rubric/(?P<rubric_component_id>[0-9]+)$', views.RubricDetail.as_view(), name="rubric-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/register', views.Register.as_view(), name="register"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/$', views.TeamList.as_view(), name="team-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)$', views.TeamDetail.as_view(), name="team-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/students/$', views.TeamMemberList.as_view(), name="teammember-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/students/(?P<student_username>[a-zA-Z0-9_-]+)$', views.TeamMemberDetail.as_view(), name="teammember-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/assignments/$', views.RegistrationList.as_view(), name="registration-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)$', views.RegistrationDetail.as_view(), name="registration-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/submit$', views.Submit.as_view(), name="submit"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/submissions/$', views.SubmissionList.as_view(), name="submission-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/submissions/(?P<submission_id>[0-9]+)$', views.SubmissionDetail.as_view(), name="submission-detail"),

    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/grades/$', views.GradeList.as_view(), name="grade-list"),
    url(URL_PREFIX + r'courses/(?P<course_id>[a-zA-Z0-9_-]+)/teams/(?P<team_id>[a-zA-Z0-9_-]+)/assignments/(?P<assignment_id>[a-zA-Z0-9_-]+)/grades/(?P<grade_id>[0-9]+)$', views.GradeDetail.as_view(), name="grade-detail"),

    url(URL_PREFIX + r'users/$', views.UserList.as_view(), name="user-list"),
    url(URL_PREFIX + r'users/(?P<username>[a-zA-Z0-9_-]+)/$', views.UserDetail.as_view(), name="user-detail"),
    url(URL_PREFIX + r'users/(?P<username>[a-zA-Z0-9_-]+)/token/$', views.UserToken.as_view(), name="user-token"),
    
    url(URL_PREFIX + r'user/$', views.AuthUserDetail.as_view(), name="auth-user-detail"),
    url(URL_PREFIX + r'user/token/$', views.AuthUserToken.as_view(), name="auth-user-token"),
]

urlpatterns = format_suffix_patterns(urlpatterns)


from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.contrib.auth.models import User

from pprint import pprint
from chisubmit.backend.api.models import Course, Student, Instructor

class CourseTests(APITestCase):
    
    fixtures = ['users', 'course1', 'course1_users']
    
    def test_get_courses(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)
        
        url = reverse('course-list')
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_get_course(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('course-detail', args=["cmsc40100"])
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_patch_course_as_admin(self):
        user = User.objects.get(username='admin')
        self.client.force_authenticate(user=user)

        url = reverse('course-detail', args=["cmsc40100"])
        response = self.client.patch(url, data={"name":"FOOBAR"})
        c = Course.objects.get(course_id='cmsc40100')
        self.assertEqual(c.name, "FOOBAR") 
        
        self.assertEqual(response.status_code, status.HTTP_200_OK)   
        
    def test_patch_course_as_student(self):
        user = User.objects.get(username='student1')
        self.client.force_authenticate(user=user)

        url = reverse('course-detail', args=["cmsc40100"])
        response = self.client.patch(url, data={"name":"FOOBAR"})
        c = Course.objects.get(course_id='cmsc40100')
        self.assertEqual(c.name, "Introduction to Software Testing") 
        
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        
    def test_patch_course_instructor(self):
        user = User.objects.get(username='instructor1')
        self.client.force_authenticate(user=user)

        url = reverse('instructor-detail', args=["cmsc40100", "instructor1"])
        response = self.client.patch(url, data={"git_username": "git-instructor1"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)           

        instructor_obj = Instructor.objects.get(course__course_id = 'cmsc40100', user__username = "instructor1")
        self.assertEquals(instructor_obj.git_username, "git-instructor1")
        
        
    
        


from chisubmit.tests.integration.clientlibs import ChisubmitClientLibsTestCase
from chisubmit.tests.common import COURSE1_ASSIGNMENTS, COURSE2_ASSIGNMENTS
from chisubmit.client.exceptions import BadRequestException
from chisubmit.backend.api.models import Course, Assignment

class AssignmentTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_pa1', 'course1_pa2']
    
    def test_get_assignments(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        assignments = course.get_assignments()
        
        self.assertEquals(len(assignments), len(COURSE1_ASSIGNMENTS))
        self.assertItemsEqual([a.assignment_id for a in assignments], COURSE1_ASSIGNMENTS)
        
    def test_get_assignment(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for assignment_id in COURSE1_ASSIGNMENTS:
            assignment = course.get_assignment(assignment_id)
            
            self.assertEquals(assignment.assignment_id, assignment_id)
            
    def test_create_assignment(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        assignment = course.create_assignment(assignment_id = "pa3",
                                              name = "Programming Assignment 3",
                                              deadline = "2042-02-04 20:00:00+00:00",
                                              min_students = 2,
                                              max_students = 2)
        self.assertEquals(assignment.assignment_id, "pa3")
        self.assertEquals(assignment.name, "Programming Assignment 3")
        
        course_obj = Course.get_by_course_id("cmsc40100")
        self.assertIsNotNone(course_obj)
        
        assignment_obj = course_obj.get_assignment("pa3")
        self.assertIsNotNone(assignment_obj, "Assignment was not added to database")
            
        self.assertEquals(assignment_obj.assignment_id, "pa3")                  
        self.assertEquals(assignment_obj.name, "Programming Assignment 3")                  
        self.assertEquals(assignment_obj.min_students, 2)                  
        self.assertEquals(assignment_obj.max_students, 2)                  
                    
            
            
class AssignmentValidationTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_pa1', 'course1_pa2',
                         'course2', 'course2_hw1', 'course2_hw2']
    
    def test_create_assignment_invalid_id(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        with self.assertRaises(BadRequestException) as cm:
            assignment = course.create_assignment(assignment_id = "pa 3",
                                                  name = "Programming Assignment 3",
                                                  deadline = "2042-02-04 20:00:00+00:00",
                                                  min_students = 2,
                                                  max_students = 2)
        
        bre = cm.exception
        self.assertItemsEqual(bre.errors.keys(), ["assignment_id"])
        self.assertEqual(len(bre.errors["assignment_id"]), 1)
        
    def test_create_assignment_multiple_errors(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        with self.assertRaises(BadRequestException) as cm:
            assignment = course.create_assignment(assignment_id = "pa 3",
                                                  name = None,
                                                  deadline = "2042-02-04 20:00:00+00:00",
                                                  min_students = 2,
                                                  max_students = 2)
        
        bre = cm.exception
        self.assertItemsEqual(bre.errors.keys(), ["assignment_id", "name"])
        self.assertEqual(len(bre.errors["assignment_id"]), 1)        
        self.assertEqual(len(bre.errors["name"]), 1)        
          
    def test_create_assignment_existing_assignment(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        with self.assertRaises(BadRequestException) as cm:
            assignment = course.create_assignment(assignment_id = "pa1",
                                                  name = "Programming Assignment 1",
                                                  deadline = "2042-02-04 20:00:00+00:00",
                                                  min_students = 2,
                                                  max_students = 2)
        
        bre = cm.exception
        self.assertItemsEqual(bre.errors.keys(), ["assignment_id"])
        self.assertEqual(len(bre.errors["assignment_id"]), 1)      
        
    def test_create_assignment_unique_within_course(self):
        c = self.get_api_client("admintoken")
        
        course1 = c.get_course("cmsc40100")
        course2 = c.get_course("cmsc40110")
        
        with self.assertRaises(BadRequestException):
            assignment1 = course1.create_assignment(assignment_id = "pa1",
                                                    name = "Programming Assignment 1",
                                                    deadline = "2042-02-04 20:00:00+00:00",
                                                    min_students = 2,
                                                    max_students = 2)        
        
        with self.assertRaises(BadRequestException):
            assignment2 = course2.create_assignment(assignment_id = "hw1",
                                                    name = "Homework 1",
                                                    deadline = "2042-02-04 20:00:00+00:00",
                                                    min_students = 2,
                                                    max_students = 2)                
        
        # Should not raise exceptions because course1 has pa1, pa2
        # and course2 has hw1, hw2
        assignment1 = course1.create_assignment(assignment_id = "hw1",
                                                name = "Homework 1",
                                                deadline = "2042-02-04 20:00:00+00:00",
                                                min_students = 2,
                                                max_students = 2)

        assignment2 = course2.create_assignment(assignment_id = "pa1",
                                                name = "Programming Assignment 1",
                                                deadline = "2042-02-04 20:00:00+00:00",
                                                min_students = 2,
                                                max_students = 2)
        
        assignments1 = course1.get_assignments()
        
        self.assertEquals(len(assignments1), len(COURSE1_ASSIGNMENTS) + 1)
        self.assertItemsEqual([a.assignment_id for a in assignments1], COURSE1_ASSIGNMENTS + ["hw1"])        
        
        assignments2 = course2.get_assignments()        
        
        self.assertEquals(len(assignments2), len(COURSE2_ASSIGNMENTS) + 1)
        self.assertItemsEqual([a.assignment_id for a in assignments2], COURSE2_ASSIGNMENTS + ["pa1"])        
                  


import click
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.client.course import Course
from chisubmit.cli.common import pass_course, get_course_or_exit,\
    api_obj_set_attribute, get_instructor_or_exit, get_grader_or_exit,\
    get_student_or_exit, catch_chisubmit_exceptions, require_config,\
    require_local_config
from chisubmit.repos.factory import RemoteRepositoryConnectionFactory


@click.command(name="list")
@catch_chisubmit_exceptions
@require_config
@click.pass_context
def shared_course_list(ctx):  
    courses = ctx.obj["client"].get_courses()
    
    for course in courses:
        print course.course_id, course.name

    return CHISUBMIT_SUCCESS


@click.command(name="get-git-credentials")
@click.option('--username', prompt='Enter your git username')
@click.option('--password', prompt='Enter your git password', hide_input=True)
@click.option('--no-save', is_flag=True)
@click.option('--delete-permissions', is_flag=True)
@click.option('--staging', is_flag=True)
@catch_chisubmit_exceptions
@pass_course
@click.pass_context
def shared_course_get_git_credentials(ctx, course, username, password, no_save, delete_permissions, staging):
    if not staging:
        connstr = course.git_server_connstr
    else:
        connstr = course.git_staging_connstr
    
    if connstr is None or connstr == "":
        print "Course '%s' doesn't seem to be configured to use a Git server." % course.id
        ctx.exit(CHISUBMIT_FAIL)
        
    conn = RemoteRepositoryConnectionFactory.create_connection(connstr, staging = staging)
    server_type = conn.get_server_type_name()

    token, existing = conn.get_credentials(username, password, delete_repo = delete_permissions)

    if token is None:
        print "Unable to create token. Incorrect username/password."
    else:
        if not no_save:
            if ctx.obj['config']['git-credentials'] is None:
                ctx.obj['config']['git-credentials'] = {}
            ctx.obj['config']['git-credentials'][server_type] = token
            ctx.obj['config'].save()
        
        if existing:
            print "Your existing %s access token is: %s" % (server_type, token)
        else:
            print "The following %s access token has been created: %s" % (server_type, token)

        if not no_save:
            print "chisubmit has been configured to use this token from now on."

    return CHISUBMIT_SUCCESS

@click.command(name="set-user-attribute")
@click.argument('user_type', type=click.Choice(["instructor", "grader", "student"]))
@click.argument('username', type=str)
@click.argument('attr_name', type=str)
@click.argument('attr_value', type=str)
@catch_chisubmit_exceptions
@require_local_config
@pass_course
@click.pass_context
def shared_course_set_user_attribute(ctx, course, user_type, username, attr_name, attr_value):
    if user_type == "instructor":
        user = get_instructor_or_exit(ctx, course, username)
    elif user_type == "grader":
        user = get_grader_or_exit(ctx, course, username)
    elif user_type == "student":
        user = get_student_or_exit(ctx, course, username)
    
    api_obj_set_attribute(ctx, user, attr_name, attr_value)



#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import os.path
import shutil
import yaml
import collections
from chisubmit.repos.factory import RemoteRepositoryConnectionFactory
from chisubmit.common import ChisubmitException


class ConfigDirectoryNotFoundException(Exception):
    pass

class Config(object):

    SYSTEM_CONFIG_FILENAME = "/etc/chisubmit/chisubmit.conf"
    GLOBAL_CONFIG_FILENAME = os.path.expanduser("~/.chisubmitconf")
    
    CONFIG_DIRNAME = ".chisubmit"
    
    SYSTEM = 0
    GLOBAL = 1
    LOCAL = 2    
    
    OPTION_API_URL = "api-url"
    OPTION_API_KEY = "api-key"
    OPTION_COURSE = "course"
    OPTION_GIT_CREDENTIALS = "git-credentials"
    
    VALID_OPTIONS = [OPTION_API_URL, OPTION_API_KEY, OPTION_COURSE, OPTION_GIT_CREDENTIALS]

    @staticmethod
    def get_config_file_values(config_file):
        if not os.path.exists(config_file):
            return {}
        
        with open(config_file, 'r') as f:
            config_file_values = yaml.safe_load(f)
    
        if type(config_file_values) != dict:
            raise ChisubmitException("{} is not valid YAML".format(f))
        
        return config_file_values
    
    @staticmethod
    def save_config_file_values(config_file, config_values):
        try:
            f = open(config_file, 'w')
            yaml.safe_dump(config_values, f, default_flow_style=False)
            f.close()
        except IOError, ioe:
            raise ChisubmitException("Error when saving configuration to file %s: %s" % (config_file, ioe.meesage), ioe)

    @staticmethod
    def set_config_value_in_file(config_file, option, value):
        config_file_values = Config.get_config_file_values(config_file)
        config_file_values[option] = value
        Config.save_config_file_values(config_file, config_file_values)

    @staticmethod
    def get_global_config_values():
        config = {}
        
        for config_file in [Config.SYSTEM_CONFIG_FILENAME, Config.GLOBAL_CONFIG_FILENAME]:
            if os.path.exists(config_file):
                config_file_values = Config.get_config_file_values(config_file)
                
                config.update(config_file_values)     
                
        return config   

    @classmethod
    def get_global_config(cls, config_overrides = {}):
        config = cls.get_global_config_values()
        config.update(config_overrides)

        return cls(config_dir = None, 
                   work_dir = None, 
                   config_values = config)


    @classmethod
    def get_config(cls, config_dir = None, work_dir = None, config_overrides = {}):
        assert( (config_dir is None and work_dir is None) or (config_dir is not None and work_dir is not None) )
        
        # If a configuration directory isn't specified, try to find it.
        if config_dir is None and work_dir is None:
            dirl = os.getcwd().split(os.sep) + [Config.CONFIG_DIRNAME]
            
            local_config_file = None
            while len(dirl) > 2:
                config_dir = os.sep.join(dirl)
                
                if os.path.exists(config_dir):
                    local_config_file = os.sep.join([config_dir, "chisubmit.conf"])
                    work_dir = os.sep.join(dirl[:-1]) 
                    break
                
                dirl.pop(-2)
            
            if local_config_file is None:
                raise ConfigDirectoryNotFoundException()
        else:
            local_config_file = os.sep.join([config_dir, "chisubmit.conf"])
                
        config = cls.get_global_config_values()
        local_config_values = cls.get_config_file_values(local_config_file)
        config.update(local_config_values)
        config.update(config_overrides)
                
        # Check for configuration values
        if not Config.OPTION_API_URL in config:
            raise ChisubmitException("Configuration value 'api-url' not found")
            
        if not Config.OPTION_API_KEY in config:
            raise ChisubmitException("No chisubmit credentials were found!")
        
        return cls(config_dir, work_dir, config)

    @classmethod
    def create_config_dir(cls, config_dir = None, work_dir = None): 
        if work_dir is None:
            work_dir = os.getcwd()
            
        if config_dir is None:
            config_dir = work_dir + "/.chisubmit"
        
        # Create directory if it does not exist
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
            
        config_values = cls.get_global_config_values()            
        
        return cls(config_dir, work_dir, config_values)

    def __init__(self, config_dir, work_dir, config_values):
        self.config_dir = config_dir
        self.work_dir = work_dir
        
        self.config_values = {opt:config_values.get(opt) for opt in Config.VALID_OPTIONS}
        
    def __get_config_file(self, where):
        if where == Config.SYSTEM:
            return Config.SYSTEM_CONFIG_FILENAME
        elif where == Config.GLOBAL:
            return Config.GLOBAL_CONFIG_FILENAME
        elif where == Config.LOCAL:
            return "{}/chisubmit.conf".format(self.config_dir)


    def get_api_url(self):
        return self.config_values[Config.OPTION_API_URL]
    
    def set_api_url(self, api_url, where = None):
        if where is None: where = Config.LOCAL
                
        config_file = self.__get_config_file(where)
        Config.set_config_value_in_file(config_file, Config.OPTION_API_URL, api_url)
        
        
    def get_api_key(self):
        return self.config_values[Config.OPTION_API_KEY]
    
    def set_api_key(self, api_key, where = None):
        if where is None: where = Config.LOCAL
                
        config_file = self.__get_config_file(where)
        Config.set_config_value_in_file(config_file, Config.OPTION_API_KEY, api_key)
        
        
    def get_course(self):
        return self.config_values[Config.OPTION_COURSE]
    
    def set_course(self, course_id, where = None):
        if where is None: where = Config.LOCAL
                
        config_file = self.__get_config_file(where)
        Config.set_config_value_in_file(config_file, Config.OPTION_COURSE, course_id)


    def get_git_credentials(self, server_type):
        if server_type not in RemoteRepositoryConnectionFactory.server_types:
            raise ValueError("Unknown git server type: {}".format(server_type))

        if self.config_values.get(Config.OPTION_GIT_CREDENTIALS) is None:
            return None
        else:
            return self.config_values[Config.OPTION_GIT_CREDENTIALS].get(server_type)
    
    def set_git_credentials(self, server_type, credentials, where = None):
        if where is None: where = Config.LOCAL
        
        if server_type not in RemoteRepositoryConnectionFactory.server_types:
            raise ValueError("Unknown git server type: {}".format(server_type))
                
        config_file = self.__get_config_file(where)
        
        config_file_values = Config.get_config_file_values(config_file)
        if not Config.OPTION_GIT_CREDENTIALS in config_file_values:
            config_file_values[Config.OPTION_GIT_CREDENTIALS] = {}
        
        config_file_values[Config.OPTION_GIT_CREDENTIALS][server_type] = credentials
        
        Config.save_config_file_values(config_file, config_file_values)


from chisubmit.tests.integration.clientlibs import ChisubmitClientLibsTestCase
from chisubmit.tests.common import COURSE1_ASSIGNMENTS, COURSE2_ASSIGNMENTS,\
    COURSE1_RUBRICS
from chisubmit.client.exceptions import BadRequestException
from chisubmit.backend.api.models import Course, Assignment

class RubricComponentTests(ChisubmitClientLibsTestCase):
    
    fixtures = ['users', 'course1', 'course1_pa1', 'course1_pa2']
    
    def test_get_rubric_components(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        
        for assignment_id in COURSE1_ASSIGNMENTS:
            assignment = course.get_assignment(assignment_id)
            rcs = assignment.get_rubric_components()
            
            self.assertEquals(len(rcs), len(COURSE1_RUBRICS[assignment_id]))
            self.assertItemsEqual([rc.description for rc in rcs], COURSE1_RUBRICS[assignment_id])
            
    def test_create_rubric_component(self):
        c = self.get_api_client("admintoken")
        
        course = c.get_course("cmsc40100")
        assignment = course.get_assignment("pa1")
        rc = assignment.create_rubric_component(description = "Third Task",
                                                points = 50,
                                                order = 3)
        
        course_obj = Course.get_by_course_id("cmsc40100")
        self.assertIsNotNone(course_obj)
        
        assignment_obj = course_obj.get_assignment("pa1")
        self.assertIsNotNone(assignment_obj)
        
        rc_obj = assignment_obj.get_rubric_component_by_description("Third Task")
        self.assertIsNotNone(rc_obj, "Rubric Component was not added to database")
            
        self.assertEquals(rc_obj.description, "Third Task")                  
        self.assertEquals(rc_obj.points, 50)                  
        self.assertEquals(rc_obj.order, 3)                  
        


import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'rest_framework.authtoken',
    'chisubmit.backend.api',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',
)

ROOT_URLCONF = 'chisubmit.backend.urls'


LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

STATIC_URL = '/static/'


REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.TokenAuthentication',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    )                  
}




from django.contrib import admin
import chisubmit.backend.api.models as models

class InstructorAdmin(admin.ModelAdmin):
    list_display = ('user', 'course')

class GraderAdmin(admin.ModelAdmin):
    list_display = ('user', 'course')
    
class StudentAdmin(admin.ModelAdmin):
    list_display = ('user', 'course', 'dropped')

class AssignmentAdmin(admin.ModelAdmin):
    list_display = ('assignment_id', 'name', 'deadline', 'course')
    
class RubricComponentAdmin(admin.ModelAdmin):
    list_display = ('assignment', 'description', 'points')
    
class TeamAdmin(admin.ModelAdmin):
    list_display = ('course', 'team_id')
    
class RegistrationAdmin(admin.ModelAdmin):
    list_display = ('team', 'assignment')      

admin.site.register(models.Course)
admin.site.register(models.Instructor, InstructorAdmin)
admin.site.register(models.Grader, GraderAdmin)
admin.site.register(models.Student, StudentAdmin)
admin.site.register(models.Assignment, AssignmentAdmin)
admin.site.register(models.RubricComponent, RubricComponentAdmin)
admin.site.register(models.Team, TeamAdmin)
admin.site.register(models.Registration, RegistrationAdmin)
admin.site.register(models.Submission)
admin.site.register(models.Grade)




from pprint import pprint

class ChisubmitRequestException(Exception):

    def __init__(self, method, url, params, data, headers, response):
        Exception.__init__(self)
        self.__method = method
        self.__url = url
        self.__params = params
        self.__data = data
        self.__headers = headers
        self.__response = response

    @property
    def method(self):
        return self.__method

    @property
    def url(self):
        return self.__url
    
    @property
    def headers(self):
        return self.__headers    
    
    @property
    def params(self):
        return self.__params    
    
    @property
    def request_data(self):
        return self.__data    

    @property
    def status(self):
        return self.__response.status_code

    @property
    def reason(self):
        return self.__response.reason

    @property
    def json(self):    
        try:
            return self.__response.json()
        except ValueError:
            return {"data": self.__response.text}  

    @property
    def response_data(self):
        return self.__response.text

    def __str__(self):
        return "HTTP %i %s (%s %s)" % (self.status, self.reason, self.method, self.url)
    
    def print_debug_info(self):
        print "HTTP REQUEST"
        print "============"
        print "%s %s" % (self.method, self.url)
        print
        print "Headers"
        print "-------"
        for hname, hvalue in self.headers.items():
            print "%s: %s" % (hname, hvalue) 
        print
        print "Query string (GET parameters)"
        print "-----------------------------"
        if self.params is None:
            print "No querystring parameters"
        else:
            for pname, pvalue in self.params.items():
                print "%s: %s" % (pname, pvalue) 
        print
        print "Request data"
        print "------------"
        if self.request_data is None:
            print "No data included in request"
        else:
            pprint(self.request_data)
        print        
        print "HTTP RESPONSE"
        print "============="
        print "%s %s" % (self.status, self.reason)
        print
        print "Headers"
        print "-------"
        for hname, hvalue in self.__response.headers.items():
            print "%s: %s" % (hname, hvalue) 
        print
        print "Response body"
        print "-------------"        
        try:
            response_data = self.__response.json()
            pprint(response_data)
        except ValueError:
            response_data = self.__response.text
            print response_data
    
    
class UnauthorizedException(ChisubmitRequestException):
    def __init__(self, *args, **kwargs):
        super(UnauthorizedException, self).__init__(*args, **kwargs) 
    
    
class UnknownObjectException(ChisubmitRequestException):
    def __init__(self, *args, **kwargs):
        super(UnknownObjectException, self).__init__(*args, **kwargs) 


class BadRequestException(ChisubmitRequestException):
    
    def __init__(self, *args, **kwargs):
        super(BadRequestException, self).__init__(*args, **kwargs)
        
        try:
            self.errors = self.json
            if any([not isinstance(k, (str, unicode)) for k in self.errors.keys()]):
                raise ValueError
            if any([not isinstance(v, (list, tuple)) for v in self.errors.values()]):
                raise ValueError
            self.valid_errors = True        
        except ValueError:
            self.errors = {}
            self.valid_errors = False
              
    def print_errors(self):
        if self.valid_errors:
            for origin, reasons in self.errors.items():
                print origin
                for r in reasons:
                    print " - %s" % r
                print
        else:
            print "HTTP 400 response included incorrectly formatted error data:"
            print self.response_data
            
